from pwnypack.target import Architecture, Endianness, Target
from pwnypack.packing import U16, U32, unpack, pack_size
from enum import Enum


__all__ = [
    'ELF',
]


class ELF(Target):
    MAGIC = b'\x7fELF'

    class Type(Enum):
        none = 0
        relocatable = 1
        executable = 2
        shared = 3
        core = 4

    class OSABI(Enum):
        system_v = 0x00
        hp_ux = 0x01
        netbsd = 0x02
        linux = 0x03
        solaris = 0x06
        aix = 0x07
        irix = 0x08
        freebsd = 0x09
        openbsd = 0x0c
        openvms = 0x0d

    class SectionType(Enum):
        UNKNOWN = -1
        NULL = 0
        PROGBITS = 1
        SYMTAB = 2
        STRTAB = 3
        RELA = 4
        HASH = 5
        DYNAMIC = 6
        NOTE = 7
        NOBITS = 8
        REL = 9
        DYNSYM = 11
        INIT_ARRAY = 14
        FINI_ARRAY = 15
        PREINIT_ARRAY = 16
        GROUP = 17
        SYMTAB_SHNDX = 18
        NUM = 19
        GNU_ATTRIBUTES = 0x6ffffff5
        GNU_HASH = 0x6ffffff6
        GNU_LIBLIST = 0x6ffffff7
        CHECKSUM = 0x6ffffff8
        LOSUNW = 0x6ffffffa
        SUNW_move = 0x6ffffffa
        SUNW_COMDAT = 0x6ffffffb
        SUNW_syminfo = 0x6ffffffc
        GNU_VERDEF = 0x6ffffffd
        GNU_VERNEED = 0x6ffffffe
        GNU_VERSYM = 0x6fffffff
        HISUNW = 0x6fffffff
        HIOS = 0x6fffffff

    # Flags for section headers.
    SHF_WRITE = 1 << 0
    SHF_ALLOC = 1 < 1
    SHF_EXECINSTR = 1 << 2
    SHF_MERGE = 1 << 4
    SHF_STRINGS = 1 << 5
    SHF_INFO_LINK = 1 << 6
    SHF_LINK_ORDER = 1 << 7
    SHF_OS_NONCONFORMING = 1 << 8
    SHF_GROUP = 1 << 9
    SHF_TLS = 1 << 10
    SHF_MASKOS = 0x0ff00000
    SHF_MASKPROC = 0xf0000000
    SHF_ORDERED = 1 << 30
    SHF_EXCLUDE = 1 << 31

    def __init__(self, header=None):
        super(ELF, self).__init__()
        self.osabi = self.abi_version = self.type = self.entry = self.phoff = \
            self.shoff = self.flags = self.hsize = self.phentsize = self.phnum = \
            self.shentsize = self.shnum = self.shstrndx = \
            self.strings = self._strings = None

        self.sections = []

        if header is not None:
            self.parse_header(header)

    def parse_header(self, data):
        (magic, bits, endian, version, osabi, abi_version, _), data = \
            unpack('4sBBBBB7s', data[:16]), data[16:]

        assert magic == self.MAGIC, 'Missing ELF magic'

        assert bits in (1, 2), 'Invalid word size'
        self.bits = 32 * bits

        self.endian = Endianness(endian)

        assert version == 1, 'Invalid version'

        self.osabi = self.OSABI(osabi)
        self.abi_version = abi_version

        (type_, arch, version), data = unpack('HHI', data[:8], endian=self.endian), data[8:]

        self.type = self.Type(type_)
        self.arch = Architecture(arch)

        assert version == 1, 'Invalid version'

        if self.bits == 32:
            format = 'IIIIHHHHHH'
        else:
            format = 'QQQIHHHHHH'

        format_size = pack_size(format)
        (self.entry, self.phoff, self.shoff, self.flags, self.hsize, self.phentsize,
            self.phnum, self.shentsize, self.shnum, self.shstrndx) = \
            unpack(format, data[:format_size], target=self)

    def parse_section_header(self, data):
        if self.bits == 32:
            format = 'I' * 10
        else:
            format = 'IIQQQQIIQQ'

        format_size = pack_size(format)
        section = {
            key: value
            for key, value in zip(
                [
                    'name_index',
                    'type_id',
                    'flags',
                    'addr',
                    'offset',
                    'size',
                    'link',
                    'info',
                    'addralign',
                    'entsize',
                ],
                unpack(format, data[:format_size], target=self)
            )
        }

        if self._strings is not None:
            name_index = section['name_index']
            name = self._strings[name_index:].split('\0', 1)[0]
            self.strings[name_index] = name
        else:
            name = None

        try:
            type_ = self.SectionType(section['type_id'])
        except ValueError:
            type_ = self.SectionType.UNKNOWN

        section.update({
            'type': type_,
            'name': name,
        })

        return section

    def parse_strings(self, data):
        self._strings = data.decode('ascii')
        self.strings = {}  # Will be filled by parsing sections.

    @classmethod
    def parse(cls, f):
        if type(f) is str:
            f = open(f, 'rb')
            need_close = True
        else:
            need_close = False

        elf = cls(f.read(64))

        if elf.shnum:
            f.seek(elf.shoff + elf.shstrndx * elf.shentsize)
            str_sh = elf.parse_section_header(f.read(elf.shentsize))

            f.seek(str_sh['offset'])
            elf.parse_strings(f.read(str_sh['size']))

            f.seek(elf.shoff)
            for i in range(elf.shnum):
                elf.sections.append(elf.parse_section_header(f.read(elf.shentsize)))

        if need_close:
            f.close()

        return elf



import pwnypack.main


__all__ = [
    'shell',
]


@pwnypack.main.register()
def shell(argparse, cmd, args):
    """
    Start an interactive IPython shell with pwny imported globally.
    """

    import pwny
    from IPython import start_ipython
    from IPython.config import get_config

    pwny_locals = {
        key: getattr(pwny, key)
        for key in dir(pwny)
        if not key.startswith('__') and not key == 'shell'
    }

    config = get_config()
    config.InteractiveShell.confirm_exit = False
    start_ipython(
        argv=args,
        config=config,
        user_ns=pwny_locals,
    )



from __future__ import print_function
from collections import OrderedDict
import io
import os
import sys
import argparse
import six


MAIN_FUNCTIONS = OrderedDict()


def register(name=None):
    def wrapper(f):
        if name is None:
            f_name = f.__name__
        else:
            f_name = name
        if f_name in MAIN_FUNCTIONS:
            raise ValueError('Duplicate application %s' % f_name)
        MAIN_FUNCTIONS[f_name] = f
        return f
    return wrapper


def binary_value_or_stdin(value):
    """
    Return fsencoded value or read raw data from stdin if value is None.
    """
    if value is None:
        reader = io.open(sys.stdin.fileno(), mode='rb', closefd=False)
        return reader.read()
    elif six.PY3:
        return os.fsencode(value)
    else:
        return value


def string_value_or_stdin(value):
    """
    Return value or read string from stdin if value is None.
    """
    if value is None:
        return sys.stdin.read()
    else:
        return value


def main():
    def usage():
        global MAIN_FUNCTIONS
        print('Welcome to pwny!')
        print()
        print('Available apps:')
        longest_app_name = max(len(app) for app in MAIN_FUNCTIONS)
        for app, f in MAIN_FUNCTIONS.items():
            fmt = ' - %%-%ds   %%s' % longest_app_name
            print(fmt % (app, f.__doc__.strip().split('\n')[0]))
        print()
        sys.exit(1)

    # Import everything so all main functions are registered.
    import pwny

    global MAIN_FUNCTIONS
    if len(sys.argv) < 2 or sys.argv[1] not in MAIN_FUNCTIONS:
        usage()
    else:
        app, app_args = sys.argv[1], sys.argv[2:]

        f = MAIN_FUNCTIONS[app]
        parser = argparse.ArgumentParser(
            prog='%s %s' % (os.path.split(sys.argv[0])[1], app),
            description=f.__doc__,
        )
        parser.add_argument(
            '-f', '--format',
            dest='format',
            choices=['raw', 'hex', 'py', 'sh', 'b64'],
            default='raw',
            help='set output format'
        )
        parser.add_argument(
            '-n', '--no-newline',
            dest='no_newline',
            action='store_const',
            const=True,
            help='inhibit newline after output'
        )

        output = f(parser, app, app_args)

        if output is not None:
            if six.PY3:
                output_bytes = os.fsencode(output)
            else:
                output_bytes = output

            args = parser.parse_args(app_args)
            if args.no_newline:
                end = ''
            else:
                end = '\n'

            if args.format == 'raw':
                writer = io.open(sys.stdout.fileno(), mode='wb', closefd=False)
                writer.write(output_bytes)
                if not args.no_newline:
                    writer.write(b'\n')
            elif args.format == 'hex':
                print(pwny.enhex(output_bytes), end=end)
            elif args.format == 'py':
                print(repr(output), end=end)
            elif args.format == 'sh':
                r = repr(output)
                if r.startswith('b\''):
                    r = r[1:]
                print('$' + r, end=end)
            elif args.format == 'b64':
                print(pwny.enb64(output_bytes), end=end)



from six.moves import range
import pwnypack.main


__all__ = [
    'cycle',
]


def deBruijn(n, k):
    '''
    An implementation of the FKM algorithm for generating the de Bruijn
    sequence containing all k-ary strings of length n, as described in
    "Combinatorial Generation" by Frank Ruskey.
    '''

    a = [ 0 ] * (n + 1)

    def gen(t, p):
        if t > n:
            for v in a[1:p + 1]:
                yield v
        else:
            a[t] = a[t - p]
         
            for v in gen(t + 1, p):
                yield v
         
            for j in range(a[t - p] + 1, k):
                a[t] = j
                for v in gen(t + 1, t):
                    yield v

    return gen(1, 1)


def cycle(length, width=4, **kwargs):
    iter = deBruijn(width, 26)
    return ''.join([chr(ord('A') + next(iter)) for i in range(length)])


def find(key, width=4):
    key_len = len(key)
    buf = ''

    iter = deBruijn(width, 26)

    for i in range(key_len):
        buf += chr(ord('A') + next(iter))

    if buf == key:
        return 0

    for i, c in enumerate(iter):
        buf = buf[1:] + chr(ord('A') + c)
        if buf == key:
            return i + 1

    return -1


cycle.find = find
del find


@pwnypack.main.register('cycle')
def cycle_app(parser, cmd, args):
    """
    Generate a de Bruijn sequence of a given length.
    """

    parser.add_argument('-w', '--width', type=int, default=4, help='the length of the cycled value')
    parser.add_argument('length', type=int, help='the cycle length to generate')
    args = parser.parse_args(args)
    return cycle(args.length, args.width)


@pwnypack.main.register('cycle-find')
def cycle_find_app(parser, cmd, args):
    """
    Find the first position of a value in a de Bruijn sequence.
    """

    parser.add_argument('-w', '--width', type=int, default=4, help='the length of the cycled value')
    parser.add_argument('value', help='the value to determine the position of')
    args = parser.parse_args(args)
    return 'Found at position: %d' % cycle.find(args.value, args.width)



import pwnypack.target
import struct


__all__ = [
    'pack',
    'unpack',
    'pack_size',
    'P',
    'p',
    'U',
    'u',
]


def pack(fmt, *args, **kwargs):
    endian = kwargs.get('endian', kwargs.get('target', pwnypack.target.target).endian)
    if fmt and fmt[0] not in '@=<>!':
        if endian is pwnypack.target.Endianness.little:
            fmt = '<' + fmt
        elif endian is pwnypack.target.Endianness.big:
            fmt = '>' + fmt
        else:
            raise NotImplementedError('Unsupported endianness: %s' % endian)
    return struct.pack(fmt, *args)


def unpack(fmt, data, **kwargs):
    endian = kwargs.get('endian', kwargs.get('target', pwnypack.target.target).endian)
    if fmt and fmt[0] not in '@=<>!':
        if endian is pwnypack.target.Endianness.little:
            fmt = '<' + fmt
        elif endian is pwnypack.target.Endianness.big:
            fmt = '>' + fmt
        else:
            raise NotImplementedError('Unsupported endianness: %s' % endian)
    return struct.unpack(fmt, data)


def pack_size(fmt, **kwargs):
    endian = kwargs.get('endian', kwargs.get('target', pwnypack.target.target).endian)
    if fmt and fmt[0] not in '@=<>!':
        if endian is pwnypack.target.Endianness.little:
            fmt = '<' + fmt
        elif endian is pwnypack.target.Endianness.big:
            fmt = '>' + fmt
        else:
            raise NotImplementedError('Unsupported endianness: %s' % endian)
    return struct.calcsize(fmt)


def _pack_closure(f, fmt):
    return lambda a, **k: f(fmt, a, **k)


def _unpack_closure(f, fmt):
    return lambda a, **k: f(fmt, a, **k)[0]


for _w, _f in ((8, 'b'), (16, 'h'), (32, 'l'), (64, 'q')):
    locals().update({
        'p%d' % _w: _pack_closure(pack, _f),
        'P%d' % _w: _pack_closure(pack, _f.upper()),
        'u%d' % _w: _unpack_closure(unpack, _f),
        'U%d' % _w: _unpack_closure(unpack, _f.upper()),
    })
    __all__.extend([
        'p%d' % _w,
        'P%d' % _w,
        'u%d' % _w,
        'U%d' % _w,
    ])
del _w, _f, _pack_closure, _unpack_closure


def P(value, **kwargs):
    bits = kwargs.get('bits', kwargs.get('target', pwnypack.target.target).bits)
    return globals()['P%d' % bits](value, **kwargs)


def p(value, **kwargs):
    bits = kwargs.get('bits', kwargs.get('target', pwnypack.target.target).bits)
    return globals()['p%d' % bits](value, **kwargs)


def U(data, **kwargs):
    bits = kwargs.get('bits', kwargs.get('target', pwnypack.target.target).bits)
    return globals()['U%d' % bits](data, **kwargs)


def u(data, **kwargs):
    bits = kwargs.get('bits', kwargs.get('target', pwnypack.target.target).bits)
    return globals()['u%d' % bits](data, **kwargs)



import subprocess
import pwnypack.target
import tempfile
import os
from enum import Enum


__all__ = [
    'asm',
]


class Asm(object):
    class Format(Enum):
        bin = 'bin'
        ith = 'ith'
        srec = 'srec'
        aout = 'aout'
        aoutb = 'aoutb'
        coff = 'coff'
        elf32 = 'elf32'
        elf64 = 'elf64'
        elfx32 = 'elfx32'
        as86 = 'as86'
        obj = 'obj'
        win32 = 'win32'
        win64 = 'win64'
        rdf = 'rdf'
        ieee = 'ieee'
        macho32 = 'macho32'
        macho64 = 'macho64'
        dbg = 'dbg'

    @classmethod
    def __call__(cls, code, fmt=Format.bin, bits=None, target=None):
        if not isinstance(fmt, cls.Format):
            fmt = cls.Format(fmt)

        if bits is None:
            if target is None:
                target = pwnypack.target.target
            bits = target.bits

        tmp = tempfile.NamedTemporaryFile(delete=False)
        tmp.write(('bits %d\n%s' % (bits, code)).encode('utf-8'))
        tmp.close()

        try:
            p = subprocess.Popen(
                [
                    'nasm',
                    '-o',
                    '/dev/stdout',
                    '-f',
                    fmt.value,
                    tmp.name,
                ],
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            stdout, stderr = p.communicate()
            if p.returncode:
                raise SyntaxError(stderr.decode('utf-8'))
            return stdout
        finally:
            os.unlink(tmp.name)

asm = Asm()



import subprocess
import sys
import socket


__all__ = [
    'ProcessChannel',
    'SocketChannel',
    'TCPSocketChannel',
    'Flow',
]


class ProcessChannel(object):
    def __init__(self, *arguments):
        self._process = subprocess.Popen(
            arguments,
            bufsize=0,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
        )

    def read(self, n):
        d = b''
        while n:
            try:
                block = self._process.stdout.read(n)
            except ValueError:
                block = None
            if not block:
                self._process.poll()
                raise EOFError('Process ended')
            d += block
            n -= len(block)
        return d

    def write(self, data):
        self._process.poll()
        if not self._process.returncode is None:
            raise EOFError('Process ended')
        return self._process.stdin.write(data)

    def close(self):
        self._process.communicate()

    def kill(self):
        self._process.kill()


class SocketChannel(object):
    def __init__(self, socket):
        self._socket = socket

    def read(self, n):
        d = b''
        while n:
            try:
                block = self._socket.recv(n)
            except socket.error:
                block = None
            if not block:
                raise EOFError('Socket closed')
            d += block
            n -= len(block)
        return d

    def write(self, data):
        while data:
            try:
                n = self._socket.send(data)
            except socket.error:
                n = None
            if not n:
                raise EOFError('Socket closed')
            data = data[n:]

    def close(self):
        self._socket.close()

    def kill(self):
        self._socket.shutdown(socket.SHUT_RDWR)
        self._socket.close()


class TCPSocketChannel(SocketChannel):
    def __init__(self, host, port):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect((host, port))
        super(TCPSocketChannel, self).__init__(s)


class Flow(object):
    def __init__(self, channel, echo=False):
        self.channel = channel
        self.echo = echo

    def read(self, n, echo=None):
        d = self.channel.read(n)
        if echo or (echo is None and self.echo):
            sys.stdout.write(d.decode('latin1'))
        return d

    def read_eof(self, echo=None):
        d = b''
        while True:
            try:
                d += self.read(1, echo)
            except EOFError:
                return d

    def read_until(self, s, echo=None):
        s_len = len(s)
        buf = self.read(s_len, echo)

        while buf[-s_len:] != s:
            buf += self.read(1, echo)

        return buf

    until = read_until

    def readlines(self, n, echo=None):
        return [
            self.until(b'\n', echo)
            for i in range(n)
        ]

    def readline(self, echo=None):
        return self.readlines(1, echo)[0]

    def write(self, data, echo=None):
        if echo or (echo is None and self.echo):
            sys.stdout.write(data.decode('latin1'))
        self.channel.write(data)

    def writelines(self, lines, echo=None):
        self.write(b'\n'.join(lines + [b'']), echo)

    def writeline(self, line='', echo=None):
        self.writelines([line], echo)

    def close(self):
        self.channel.close()

    def kill(self):
        self.channel.kill()

    @classmethod
    def execute(cls, *args, **kwargs):
        echo = kwargs.pop('echo', False)
        return cls(ProcessChannel(*args), echo=echo)

    @classmethod
    def connect_tcp(cls, *args, **kwargs):
        echo = kwargs.pop('echo', False)
        return cls(TCPSocketChannel(*args, **kwargs), echo=echo)



import base64
import collections
import string
import six
import codecs
import pwnypack.main


__all__ = [
    'xor',
    'rot13',
    'caesar',
    'enhex',
    'dehex',
    'enb64',
    'deb64',
    'frequency',
]


def xor(key, data):
    if type(key) is int:
        key = six.int2byte(key)
    key_len = len(key)

    return b''.join(
        six.int2byte(c ^ six.indexbytes(key, i % key_len))
        for i, c in enumerate(six.iterbytes(data))
    )


def caesar(shift, data, shift_ranges=('az', 'AZ')):
    alphabet = {
        chr(c): chr((c - s + shift) % (e - s + 1) + s)
        for s, e in map(lambda r: (ord(r[0]), ord(r[-1])), shift_ranges)
        for c in range(s, e + 1)
    }
    return ''.join(alphabet.get(c, c) for c in data)


rot13_encode = codecs.getencoder('rot-13')
rot13 = lambda d: rot13_encode(d)[0]

hex_encode = codecs.getencoder('hex')
hex_decode = codecs.getdecoder('hex')
enhex = lambda d: hex_encode(d)[0].decode('ascii')
dehex = lambda d: hex_decode(''.join(d.replace(':', '').split()))[0]


enb64 = lambda d: base64.b64encode(d).decode('ascii')
deb64 = lambda d: base64.b64decode(d.encode('ascii'))


frequency = collections.Counter


@pwnypack.main.register('xor')
def xor_app(parser, cmd, args):
    """
    Xor a value with a key.
    """

    parser.add_argument(
        '-d', '--dec',
        help='interpret the key as a decimal integer',
        dest='type',
        action='store_const',
        const=int
    )
    parser.add_argument(
        '-x', '--hex',
        help='interpret the key as an hexadecimal integer',
        dest='type',
        action='store_const',
        const=lambda v: int(v, 16)
    )
    parser.add_argument('key', help='the key to xor the value with')
    parser.add_argument('value', help='the value to xor, read from stdin if omitted', nargs='?')

    args = parser.parse_args(args)
    if args.type is not None:
        args.key = args.type(args.key)

    return xor(args.key, pwnypack.main.binary_value_or_stdin(args.value))


@pwnypack.main.register('caesar')
def caesar_app(parser, cmd, args):
    """
    Caesar crypt a value with a key.
    """

    parser.add_argument('shift', type=int, help='the shift to apply')
    parser.add_argument('value', help='the value to caesar crypt, read from stdin if omitted', nargs='?')
    parser.add_argument(
        '-s', '--shift-range',
        dest='shift_ranges',
        action='append',
        help='specify a character range to shift (defaults to a-z, A-Z)'
    )

    args = parser.parse_args(args)
    if not args.shift_ranges:
        args.shift_ranges = ['az', 'AZ']

    return caesar(args.shift, pwnypack.main.string_value_or_stdin(args.value), args.shift_ranges)


@pwnypack.main.register('rot13')
def rot13_app(parser, cmd, args):
    """
    rot13 encrypt a value.
    """

    parser.add_argument('value', help='the value to rot13, read from stdin if omitted', nargs='?')
    args = parser.parse_args(args)
    return rot13(pwnypack.main.string_value_or_stdin(args.value))


@pwnypack.main.register('enb64')
def enb64_app(parser, cmd, args):
    """
    base64 encode a value.
    """

    parser.add_argument('value', help='the value to base64 encode, read from stdin if omitted', nargs='?')
    args = parser.parse_args(args)
    return enb64(pwnypack.main.binary_value_or_stdin(args.value))


@pwnypack.main.register('deb64')
def deb64_app(parser, cmd, args):
    """
    base64 decode a value.
    """

    parser.add_argument('value', help='the value to base64 decode, read from stdin if omitted', nargs='?')
    args = parser.parse_args(args)
    return deb64(pwnypack.main.string_value_or_stdin(args.value))


@pwnypack.main.register('enhex')
def enhex_app(parser, cmd, args):
    """
    hex encode a value.
    """

    parser.add_argument('value', help='the value to hex encode, read from stdin if omitted', nargs='?')
    args = parser.parse_args(args)
    return enhex(pwnypack.main.binary_value_or_stdin(args.value))


@pwnypack.main.register('dehex')
def dehex_app(parser, cmd, args):
    """
    hex decode a value.
    """

    parser.add_argument('value', help='the value to base64 decode, read from stdin if omitted', nargs='?')
    args = parser.parse_args(args)
    return dehex(pwnypack.main.string_value_or_stdin(args.value))


@pwnypack.main.register('frequency')
def frequency_app(parser, cmd, args):
    """
    perform frequency analysis on a value.
    """

    parser.add_argument('value', help='the value to analyse, read from stdin if omitted', nargs='?')
    args = parser.parse_args(args)
    data = frequency(pwnypack.main.binary_value_or_stdin(args.value))
    return '\n'.join(
        '0x%02x (%c): %d' % (key, chr(key), value)
        if chr(key) in string.printable else
        '0x%02x ---: %d' % (key, value)
        for key, value in data.items()
    )



from enum import Enum
import platform


__all__ = [
    'Architecture',
    'Endianness',
    'Target',
    'target',
]


class Architecture(Enum):
    # Note: it is no coincidence that these values map to the e_machine ELF field.
    sparc = 0x02
    x86 = 0x03
    mips = 0x08
    powerpc = 0x14
    arm = 0x28
    superh = 0x2a
    ia_64 = 0x32
    x86_64 = 0x3e
    aarch64 = 0xb7


class Endianness(Enum):
    little = 1
    big = 2


class Target(object):
    _arch = None
    _endian = None
    _bits = None

    def __init__(self, arch=None, endian=None, bits=None):
        self.arch = arch
        self.endian = endian
        self.bits = bits

    @property
    def arch(self):
        if self._arch is None:
            if platform.architecture()[0] == '64bit':
                return Architecture.x86_64
            else:
                return Architecture.x86
        return self._arch

    @arch.setter
    def arch(self, arch):
        self._arch = arch

    @property
    def endian(self):
        if self._endian is None:
            # Both x86 and x86_64 use little endian.
            return Endianness.little
        else:
            return self._endian

    @endian.setter
    def endian(self, endian):
        self._endian = endian

    @property
    def bits(self):
        if self._bits is None:
            arch = self.arch
            if arch is Architecture.x86:
                return 32
            elif arch is Architecture.x86_64:
                return 64
            else:
                raise NotImplementedError('Unsupported architecture: %s' % arch)
        else:
            return self._bits

    @bits.setter
    def bits(self, value):
        self._bits = value

    def assume(self, target):
        self._arch = target._arch
        self._endian = target._endian
        self._bits = target._bits


target = Target()
